@@ -5,11 +5,19 @@ from typing import Any, Callable, Generic, Iterator, List, Optional, Set, TypeVa
 
 import pytest
 import pytz
-from annotated_types import BaseMetadata, GroupedMetadata, Gt, Lt, Predicate
+from annotated_types import BaseMetadata, GroupedMetadata, Gt, Lt, Not, Predicate
 from pydantic_core import CoreSchema, PydanticUndefined, core_schema
 from typing_extensions import Annotated
 
-from pydantic import BaseModel, Field, GetCoreSchemaHandler, PydanticUserError, TypeAdapter, ValidationError
+from pydantic import (
+    BaseModel,
+    BeforeValidator,
+    Field,
+    GetCoreSchemaHandler,
+    PydanticUserError,
+    TypeAdapter,
+    ValidationError,
+)
 from pydantic.errors import PydanticSchemaGenerationError
 from pydantic.fields import PrivateAttr
 from pydantic.functional_validators import AfterValidator
@@ -394,6 +402,23 @@ def test_predicate_error_python() -> None:
     ]
 
 
+def test_not_operation_error_python() -> None:
+    ta = TypeAdapter(Annotated[int, Not(lambda x: x > 5)])
+
+    with pytest.raises(ValidationError) as exc_info:
+        ta.validate_python(6)
+
+    # insert_assert(exc_info.value.errors(include_url=False))
+    assert exc_info.value.errors(include_url=False) == [
+        {
+            'type': 'not_operation_failed',
+            'loc': (),
+            'msg': 'Not of test_not_operation_error_python.<locals>.<lambda> failed',
+            'input': 6,
+        }
+    ]
+
+
 def test_annotated_field_info_not_lost_from_forwardref():
     from pydantic import BaseModel
 
@@ -424,18 +449,20 @@ def test_annotated_private_field_with_default():
     class AnnotatedPrivateFieldModel(BaseModel):
         _foo: Annotated[int, PrivateAttr(default=1)]
         _bar: Annotated[str, 'hello']
+        _baz: 'Annotated[str, PrivateAttr(default=2)]'
 
     model = AnnotatedPrivateFieldModel()
     assert model._foo == 1
+    assert model._baz == 2
 
-    assert model.__pydantic_private__ == {'_foo': 1}
+    assert model.__pydantic_private__ == {'_foo': 1, '_baz': 2}
 
     with pytest.raises(AttributeError):
         assert model._bar
 
     model._bar = 'world'
     assert model._bar == 'world'
-    assert model.__pydantic_private__ == {'_foo': 1, '_bar': 'world'}
+    assert model.__pydantic_private__ == {'_foo': 1, '_bar': 'world', '_baz': 2}
 
     with pytest.raises(AttributeError):
         assert model.bar
@@ -598,3 +625,16 @@ def test_utcoffset_validator_example_pattern() -> None:
     ta = TypeAdapter(Annotated[dt.datetime, MyDatetimeValidator(0, 4)])
     with pytest.raises(Exception):
         ta.validate_python(dt.datetime.now(pytz.timezone(LA)))
+
+
+def test_incompatible_metadata_error() -> None:
+    ta = TypeAdapter(Annotated[List[int], Field(pattern='abc')])
+    with pytest.raises(TypeError, match="Unable to apply constraint 'pattern'"):
+        ta.validate_python([1, 2, 3])
+
+
+def test_compatible_metadata_raises_correct_validation_error() -> None:
+    """Using a no-op before validator to ensure that constraint is applied as part of a chain."""
+    ta = TypeAdapter(Annotated[str, BeforeValidator(lambda x: x), Field(pattern='abc')])
+    with pytest.raises(ValidationError, match="String should match pattern 'abc'"):
+        ta.validate_python('def')
