@@ -1,5 +1,6 @@
 import re
 import sys
+from dataclasses import dataclass
 from enum import Enum, IntEnum
 from types import SimpleNamespace
 from typing import Any, Callable, Generic, List, Optional, Sequence, TypeVar, Union
@@ -7,12 +8,23 @@ from typing import Any, Callable, Generic, List, Optional, Sequence, TypeVar, Un
 import pytest
 from dirty_equals import HasRepr, IsStr
 from pydantic_core import SchemaValidator, core_schema
-from typing_extensions import Annotated, Literal
+from typing_extensions import Annotated, Literal, TypedDict
 
-from pydantic import BaseModel, ConfigDict, Discriminator, Field, TypeAdapter, ValidationError, field_validator
+from pydantic import (
+    BaseModel,
+    ConfigDict,
+    Discriminator,
+    Field,
+    PlainSerializer,
+    TypeAdapter,
+    ValidationError,
+    field_validator,
+)
 from pydantic._internal._discriminated_union import apply_discriminator
+from pydantic.dataclasses import dataclass as pydantic_dataclass
 from pydantic.errors import PydanticUserError
 from pydantic.fields import FieldInfo
+from pydantic.functional_validators import model_validator
 from pydantic.json_schema import GenerateJsonSchema
 from pydantic.types import Tag
 
@@ -973,60 +985,54 @@ def test_lax_or_strict_definitions() -> None:
     discriminated_schema = apply_discriminator(core_schema.union_schema([cat, dog]), 'kind')
     # insert_assert(discriminated_schema)
     assert discriminated_schema == {
-        'type': 'definitions',
-        'schema': {
-            'type': 'tagged-union',
-            'choices': {
-                'cat': {
+        'type': 'tagged-union',
+        'choices': {
+            'cat': {
+                'type': 'typed-dict',
+                'fields': {'kind': {'type': 'typed-dict-field', 'schema': {'type': 'literal', 'expected': ['cat']}}},
+            },
+            'DOG': {
+                'type': 'lax-or-strict',
+                'lax_schema': {
                     'type': 'typed-dict',
                     'fields': {
-                        'kind': {'type': 'typed-dict-field', 'schema': {'type': 'literal', 'expected': ['cat']}}
+                        'kind': {'type': 'typed-dict-field', 'schema': {'type': 'literal', 'expected': ['DOG']}}
                     },
                 },
-                'DOG': {
-                    'type': 'lax-or-strict',
-                    'lax_schema': {
+                'strict_schema': {
+                    'type': 'definitions',
+                    'schema': {
                         'type': 'typed-dict',
                         'fields': {
-                            'kind': {'type': 'typed-dict-field', 'schema': {'type': 'literal', 'expected': ['DOG']}}
+                            'kind': {'type': 'typed-dict-field', 'schema': {'type': 'literal', 'expected': ['dog']}}
                         },
                     },
-                    'strict_schema': {
-                        'type': 'definitions',
-                        'schema': {
-                            'type': 'typed-dict',
-                            'fields': {
-                                'kind': {'type': 'typed-dict-field', 'schema': {'type': 'literal', 'expected': ['dog']}}
-                            },
-                        },
-                        'definitions': [{'type': 'int', 'ref': 'my-int-definition'}],
+                    'definitions': [{'type': 'int', 'ref': 'my-int-definition'}],
+                },
+            },
+            'dog': {
+                'type': 'lax-or-strict',
+                'lax_schema': {
+                    'type': 'typed-dict',
+                    'fields': {
+                        'kind': {'type': 'typed-dict-field', 'schema': {'type': 'literal', 'expected': ['DOG']}}
                     },
                 },
-                'dog': {
-                    'type': 'lax-or-strict',
-                    'lax_schema': {
+                'strict_schema': {
+                    'type': 'definitions',
+                    'schema': {
                         'type': 'typed-dict',
                         'fields': {
-                            'kind': {'type': 'typed-dict-field', 'schema': {'type': 'literal', 'expected': ['DOG']}}
-                        },
-                    },
-                    'strict_schema': {
-                        'type': 'definitions',
-                        'schema': {
-                            'type': 'typed-dict',
-                            'fields': {
-                                'kind': {'type': 'typed-dict-field', 'schema': {'type': 'literal', 'expected': ['dog']}}
-                            },
+                            'kind': {'type': 'typed-dict-field', 'schema': {'type': 'literal', 'expected': ['dog']}}
                         },
-                        'definitions': [{'type': 'int', 'ref': 'my-int-definition'}],
                     },
+                    'definitions': [{'type': 'int', 'ref': 'my-int-definition'}],
                 },
             },
-            'discriminator': 'kind',
-            'strict': False,
-            'from_attributes': True,
         },
-        'definitions': [{'type': 'str', 'ref': 'my-str-definition'}],
+        'discriminator': 'kind',
+        'strict': False,
+        'from_attributes': True,
     }
 
 
@@ -1234,16 +1240,28 @@ def test_union_in_submodel() -> None:
             },
             'UnionModel1': {
                 'properties': {
-                    'type': {'const': 1, 'default': 1, 'title': 'Type'},
-                    'other': {'const': 'UnionModel1', 'default': 'UnionModel1', 'title': 'Other'},
+                    'type': {'const': 1, 'default': 1, 'enum': [1], 'title': 'Type', 'type': 'integer'},
+                    'other': {
+                        'const': 'UnionModel1',
+                        'default': 'UnionModel1',
+                        'enum': ['UnionModel1'],
+                        'title': 'Other',
+                        'type': 'string',
+                    },
                 },
                 'title': 'UnionModel1',
                 'type': 'object',
             },
             'UnionModel2': {
                 'properties': {
-                    'type': {'const': 2, 'default': 2, 'title': 'Type'},
-                    'other': {'const': 'UnionModel2', 'default': 'UnionModel2', 'title': 'Other'},
+                    'type': {'const': 2, 'default': 2, 'enum': [2], 'title': 'Type', 'type': 'integer'},
+                    'other': {
+                        'const': 'UnionModel2',
+                        'default': 'UnionModel2',
+                        'enum': ['UnionModel2'],
+                        'title': 'Other',
+                        'type': 'string',
+                    },
                 },
                 'title': 'UnionModel2',
                 'type': 'object',
@@ -1316,11 +1334,12 @@ def test_sequence_discriminated_union():
         pet: Sequence[Pet]
         n: int
 
+    # insert_assert(Model.model_json_schema())
     assert Model.model_json_schema() == {
         '$defs': {
             'Cat': {
                 'properties': {
-                    'pet_type': {'const': 'cat', 'title': 'Pet Type'},
+                    'pet_type': {'const': 'cat', 'enum': ['cat'], 'title': 'Pet Type', 'type': 'string'},
                     'meows': {'title': 'Meows', 'type': 'integer'},
                 },
                 'required': ['pet_type', 'meows'],
@@ -1329,7 +1348,7 @@ def test_sequence_discriminated_union():
             },
             'Dog': {
                 'properties': {
-                    'pet_type': {'const': 'dog', 'title': 'Pet Type'},
+                    'pet_type': {'const': 'dog', 'enum': ['dog'], 'title': 'Pet Type', 'type': 'string'},
                     'barks': {'title': 'Barks', 'type': 'number'},
                 },
                 'required': ['pet_type', 'barks'],
@@ -1371,6 +1390,90 @@ def test_sequence_discriminated_union():
     }
 
 
+def test_sequence_discriminated_union_validation():
+    """
+    Related issue: https://github.com/pydantic/pydantic/issues/9872
+    """
+
+    class A(BaseModel):
+        type: Literal['a']
+        a_field: str
+
+    class B(BaseModel):
+        type: Literal['b']
+        b_field: str
+
+    class Model(BaseModel):
+        items: Sequence[Annotated[Union[A, B], Field(discriminator='type')]]
+
+    import json
+
+    data_json = '{"items": [{"type": "b"}]}'
+    data_dict = json.loads(data_json)
+
+    expected_error = {
+        'type': 'missing',
+        'loc': ('items', 0, 'b', 'b_field'),
+        'msg': 'Field required',
+        'input': {'type': 'b'},
+    }
+
+    # missing field should be `b_field` only, not including `a_field`
+    # also `literal_error` should not be reported on `type`
+
+    with pytest.raises(ValidationError) as exc_info:
+        Model.model_validate(data_dict)
+    assert exc_info.value.errors(include_url=False) == [expected_error]
+
+    with pytest.raises(ValidationError) as exc_info:
+        Model.model_validate_json(data_json)
+    assert exc_info.value.errors(include_url=False) == [expected_error]
+
+
+def test_sequence_discriminated_union_validation_with_validator():
+    """
+    This is the same as the previous test, but add validators to both class.
+    """
+
+    class A(BaseModel):
+        type: Literal['a']
+        a_field: str
+
+        @model_validator(mode='after')
+        def check_a(self):
+            return self
+
+    class B(BaseModel):
+        type: Literal['b']
+        b_field: str
+
+        @model_validator(mode='after')
+        def check_b(self):
+            return self
+
+    class Model(BaseModel):
+        items: Sequence[Annotated[Union[A, B], Field(discriminator='type')]]
+
+    import json
+
+    data_json = '{"items": [{"type": "b"}]}'
+    data_dict = json.loads(data_json)
+
+    expected_error = {
+        'type': 'missing',
+        'loc': ('items', 0, 'b', 'b_field'),
+        'msg': 'Field required',
+        'input': {'type': 'b'},
+    }
+
+    # missing field should be `b_field` only, not including `a_field`
+    # also `literal_error` should not be reported on `type`
+
+    with pytest.raises(ValidationError) as exc_info:
+        Model.model_validate(data_dict)
+    assert exc_info.value.errors(include_url=False) == [expected_error]
+
+
 @pytest.fixture(scope='session', name='animals')
 def callable_discriminated_union_animals() -> SimpleNamespace:
     class Cat(BaseModel):
@@ -1795,12 +1898,13 @@ def test_nested_discriminator() -> None:
         blending: float
 
     MyModel.model_rebuild()
+    # insert_assert(MyModel.model_json_schema())
     assert MyModel.model_json_schema() == {
         '$defs': {
             'Step_A': {
                 'properties': {
                     'count': {'title': 'Count', 'type': 'integer'},
-                    'type': {'const': 'stepA', 'title': 'Type'},
+                    'type': {'const': 'stepA', 'enum': ['stepA'], 'title': 'Type', 'type': 'string'},
                 },
                 'required': ['type', 'count'],
                 'title': 'Step_A',
@@ -1808,7 +1912,7 @@ def test_nested_discriminator() -> None:
             },
             'Step_B': {
                 'properties': {
-                    'type': {'const': 'stepB', 'title': 'Type'},
+                    'type': {'const': 'stepB', 'enum': ['stepB'], 'title': 'Type', 'type': 'string'},
                     'value': {'title': 'Value', 'type': 'number'},
                 },
                 'required': ['type', 'value'],
@@ -1828,7 +1932,7 @@ def test_nested_discriminator() -> None:
                         'title': 'Steps',
                     },
                     'sub_models': {'items': {'$ref': '#/$defs/SubModel'}, 'title': 'Sub Models', 'type': 'array'},
-                    'type': {'const': 'mixed', 'title': 'Type'},
+                    'type': {'const': 'mixed', 'enum': ['mixed'], 'title': 'Type', 'type': 'string'},
                 },
                 'required': ['type', 'sub_models', 'blending'],
                 'title': 'SubModel',
@@ -1846,9 +1950,225 @@ def test_nested_discriminator() -> None:
                 'title': 'Steps',
             },
             'sub_models': {'items': {'$ref': '#/$defs/SubModel'}, 'title': 'Sub Models', 'type': 'array'},
-            'type': {'const': 'mixed', 'title': 'Type'},
+            'type': {'const': 'mixed', 'enum': ['mixed'], 'title': 'Type', 'type': 'string'},
         },
         'required': ['type', 'sub_models'],
         'title': 'MyModel',
         'type': 'object',
     }
+
+
+def test_nested_schema_gen_uses_tagged_union_in_ref() -> None:
+    class NestedState(BaseModel):
+        state_type: Literal['nested']
+        substate: 'AnyState'
+
+    # If this type is left out, the model behaves normally again
+    class LoopState(BaseModel):
+        state_type: Literal['loop']
+        substate: 'AnyState'
+
+    class LeafState(BaseModel):
+        state_type: Literal['leaf']
+
+    AnyState = Annotated[Union[NestedState, LoopState, LeafState], Field(..., discriminator='state_type')]
+    adapter = TypeAdapter(AnyState)
+
+    assert adapter.core_schema['schema']['type'] == 'tagged-union'
+    for definition in adapter.core_schema['definitions']:
+        if definition['schema']['model_name'] in ['NestedState', 'LoopState']:
+            assert definition['schema']['fields']['substate']['schema']['schema']['type'] == 'tagged-union'
+
+
+def test_recursive_discriminiated_union_with_typed_dict() -> None:
+    class Foo(TypedDict):
+        type: Literal['foo']
+        x: 'Foobar'
+
+    class Bar(TypedDict):
+        type: Literal['bar']
+
+    Foobar = Annotated[Union[Foo, Bar], Field(discriminator='type')]
+    ta = TypeAdapter(Foobar)
+
+    # len of errors should be 1 for each case, bc we're using a tagged union
+    with pytest.raises(ValidationError) as e:
+        ta.validate_python({'type': 'wrong'})
+    assert len(e.value.errors()) == 1
+
+    with pytest.raises(ValidationError) as e:
+        ta.validate_python({'type': 'foo', 'x': {'type': 'wrong'}})
+    assert len(e.value.errors()) == 1
+
+    core_schema = ta.core_schema
+    assert core_schema['schema']['type'] == 'tagged-union'
+    for definition in core_schema['definitions']:
+        if 'Foo' in definition['ref']:
+            assert definition['fields']['x']['schema']['type'] == 'tagged-union'
+
+
+def test_recursive_discriminiated_union_with_base_model() -> None:
+    class Foo(BaseModel):
+        type: Literal['foo']
+        x: 'Foobar'
+
+    class Bar(BaseModel):
+        type: Literal['bar']
+
+    Foobar = Annotated[Union[Foo, Bar], Field(discriminator='type')]
+    ta = TypeAdapter(Foobar)
+
+    # len of errors should be 1 for each case, bc we're using a tagged union
+    with pytest.raises(ValidationError) as e:
+        ta.validate_python({'type': 'wrong'})
+    assert len(e.value.errors()) == 1
+
+    with pytest.raises(ValidationError) as e:
+        ta.validate_python({'type': 'foo', 'x': {'type': 'wrong'}})
+    assert len(e.value.errors()) == 1
+
+    core_schema = ta.core_schema
+    assert core_schema['schema']['type'] == 'tagged-union'
+    for definition in core_schema['definitions']:
+        if 'Foo' in definition['ref']:
+            assert definition['schema']['fields']['x']['schema']['type'] == 'tagged-union'
+
+
+def test_recursive_discriminated_union_with_pydantic_dataclass() -> None:
+    @pydantic_dataclass
+    class Foo:
+        type: Literal['foo']
+        x: 'Foobar'
+
+    @pydantic_dataclass
+    class Bar:
+        type: Literal['bar']
+
+    Foobar = Annotated[Union[Foo, Bar], Field(discriminator='type')]
+    ta = TypeAdapter(Foobar)
+
+    # len of errors should be 1 for each case, bc we're using a tagged union
+    with pytest.raises(ValidationError) as e:
+        ta.validate_python({'type': 'wrong'})
+    assert len(e.value.errors()) == 1
+
+    with pytest.raises(ValidationError) as e:
+        ta.validate_python({'type': 'foo', 'x': {'type': 'wrong'}})
+    assert len(e.value.errors()) == 1
+
+    core_schema = ta.core_schema
+    assert core_schema['schema']['type'] == 'tagged-union'
+    for definition in core_schema['definitions']:
+        if 'Foo' in definition['ref']:
+            for field in definition['schema']['fields']:
+                assert field['schema']['type'] == 'tagged-union' if field['name'] == 'x' else True
+
+
+def test_discriminated_union_with_nested_dataclass() -> None:
+    @pydantic_dataclass
+    class Cat:
+        type: Literal['cat'] = 'cat'
+
+    @pydantic_dataclass
+    class Dog:
+        type: Literal['dog'] = 'dog'
+
+    @pydantic_dataclass
+    class NestedDataClass:
+        animal: Annotated[Union[Cat, Dog], Discriminator('type')]
+
+    @pydantic_dataclass
+    class Root:
+        data_class: NestedDataClass
+
+    ta = TypeAdapter(Root)
+    assert ta.core_schema['schema']['fields'][0]['schema']['schema']['fields'][0]['schema']['type'] == 'tagged-union'
+
+
+def test_discriminated_union_with_nested_typed_dicts() -> None:
+    class Cat(TypedDict):
+        type: Literal['cat']
+
+    class Dog(TypedDict):
+        type: Literal['dog']
+
+    class NestedTypedDict(TypedDict):
+        animal: Annotated[Union[Cat, Dog], Discriminator('type')]
+
+    class Root(TypedDict):
+        data_class: NestedTypedDict
+
+    ta = TypeAdapter(Root)
+    assert ta.core_schema['fields']['data_class']['schema']['fields']['animal']['schema']['type'] == 'tagged-union'
+
+
+def test_discriminated_union_with_unsubstituted_type_var() -> None:
+    T = TypeVar('T')
+
+    class Dog(BaseModel, Generic[T]):
+        type_: Literal['dog']
+        friends: List['GenericPet']
+        id: T
+
+    class Cat(BaseModel, Generic[T]):
+        type_: Literal['cat']
+        friends: List['GenericPet']
+        id: T
+
+    GenericPet = Annotated[Union[Dog[T], Cat[T]], Field(..., discriminator='type_')]
+
+    ta = TypeAdapter(Dog[int])
+    int_dog = {
+        'type_': 'dog',
+        'friends': [{'type_': 'dog', 'friends': [], 'id': 2}, {'type_': 'cat', 'friends': [], 'id': 3}],
+        'id': 1,
+    }
+    assert ta.validate_python(int_dog).id == 1
+    assert ta.validate_python(int_dog).friends[0].id == 2
+    assert ta.validate_python(int_dog).friends[1].id == 3
+
+
+def test_discriminated_union_model_dump_with_nested_class() -> None:
+    class SomeEnum(str, Enum):
+        CAT = 'cat'
+        DOG = 'dog'
+
+    class Dog(BaseModel):
+        type: Literal[SomeEnum.DOG] = SomeEnum.DOG
+        name: str
+
+    class Cat(BaseModel):
+        type: Literal[SomeEnum.CAT] = SomeEnum.CAT
+        name: str
+
+    class Yard(BaseModel):
+        pet: Union[Dog, Cat] = Field(discriminator='type')
+
+    yard = Yard(pet=Dog(name='Rex'))
+    yard_dict = yard.model_dump(mode='json')
+    assert isinstance(yard_dict['pet']['type'], str)
+    assert not isinstance(yard_dict['pet']['type'], SomeEnum)
+    assert str(yard_dict['pet']['type']) == 'dog'
+
+
+@pytest.mark.xfail(reason='Waiting for union serialization fixes via https://github.com/pydantic/pydantic/issues/9688.')
+def test_discriminated_union_serializer() -> None:
+    """Reported via https://github.com/pydantic/pydantic/issues/9590."""
+
+    @dataclass
+    class FooId:
+        _id: int
+
+    @dataclass
+    class BarId:
+        _id: int
+
+    FooOrBarId = Annotated[
+        Annotated[FooId, PlainSerializer(lambda v: {'tag': 'foo', '_id': v._id}), Tag('foo')]
+        | Annotated[BarId, PlainSerializer(lambda v: {'tag': 'bar', '_id': v._id}), Tag('bar')],
+        Discriminator(lambda v: v['tag']),
+    ]
+
+    adapter = TypeAdapter(FooOrBarId)
+    assert adapter.dump_python(FooId(1)) == {'tag': 'foo', '_id': 1}
+    assert adapter.dump_python(BarId(2)) == {'tag': 'bar', '_id': 2}
