@@ -44,6 +44,7 @@ from pydantic import (
     root_validator,
     validator,
 )
+from pydantic._internal._model_construction import ModelMetaclass
 from pydantic.fields import Field, computed_field
 from pydantic.functional_serializers import (
     field_serializer,
@@ -496,17 +497,24 @@ def test_recursive_list_error():
 
 def test_list_unions():
     class Model(BaseModel):
-        v: List[Union[int, str]] = ...
+        v: List[Union[int, str]]
 
     assert Model(v=[123, '456', 'foobar']).v == [123, '456', 'foobar']
 
     with pytest.raises(ValidationError) as exc_info:
         Model(v=[1, 2, None])
 
-    assert exc_info.value.errors(include_url=False) == [
+    # the reason that we need to do an unordered list comparison here is that previous tests use Union[str, int]
+    # and Python's cache makes it such that the above Model has `v` associated with a List[Union[str, int]] instead
+    # of the expected List[Union[int, str]]
+    # for more info, see  https://github.com/python/cpython/issues/103749 and
+    # https://github.com/pydantic/pydantic/pull/10244#issuecomment-2312796647
+    errors = exc_info.value.errors(include_url=False)
+    expected_errors = [
         {'input': None, 'loc': ('v', 2, 'int'), 'msg': 'Input should be a valid integer', 'type': 'int_type'},
         {'input': None, 'loc': ('v', 2, 'str'), 'msg': 'Input should be a valid string', 'type': 'string_type'},
     ]
+    assert sorted(errors, key=str) == sorted(expected_errors, key=str)
 
 
 def test_recursive_lists():
@@ -1387,6 +1395,26 @@ def test_type_on_annotation():
     assert Model.model_fields.keys() == set('abcdefg')
 
 
+def test_annotated_inside_type():
+    class Model(BaseModel):
+        a: Type[Annotated[int, ...]]
+
+    Model(a=int)
+
+    with pytest.raises(ValidationError) as exc_info:
+        Model(a=str)
+
+    assert exc_info.value.errors(include_url=False) == [
+        {
+            'type': 'is_subclass_of',
+            'loc': ('a',),
+            'msg': 'Input should be a subclass of int',
+            'input': str,
+            'ctx': {'class': 'int'},
+        }
+    ]
+
+
 def test_assign_type():
     class Parent:
         def echo(self):
@@ -2665,11 +2693,12 @@ def test_recursive_root_models_in_discriminated_union():
     validated = Outer.model_validate({'a': {'kind': '1', 'two': None}, 'b': {'kind': '2', 'one': None}})
     assert validated == Outer(a=Root1(root=Model1(two=None)), b=Root2(root=Model2(one=None)))
 
+    # insert_assert(Outer.model_json_schema())
     assert Outer.model_json_schema() == {
         '$defs': {
             'Model1': {
                 'properties': {
-                    'kind': {'const': '1', 'default': '1', 'title': 'Kind'},
+                    'kind': {'const': '1', 'default': '1', 'enum': ['1'], 'title': 'Kind', 'type': 'string'},
                     'two': {'anyOf': [{'$ref': '#/$defs/Model2'}, {'type': 'null'}]},
                 },
                 'required': ['two'],
@@ -2678,15 +2707,15 @@ def test_recursive_root_models_in_discriminated_union():
             },
             'Model2': {
                 'properties': {
-                    'kind': {'const': '2', 'default': '2', 'title': 'Kind'},
+                    'kind': {'const': '2', 'default': '2', 'enum': ['2'], 'title': 'Kind', 'type': 'string'},
                     'one': {'anyOf': [{'$ref': '#/$defs/Model1'}, {'type': 'null'}]},
                 },
                 'required': ['one'],
                 'title': 'Model2',
                 'type': 'object',
             },
-            'Root1': {'allOf': [{'$ref': '#/$defs/Model1'}], 'title': 'Root1'},
-            'Root2': {'allOf': [{'$ref': '#/$defs/Model2'}], 'title': 'Root2'},
+            'Root1': {'$ref': '#/$defs/Model1', 'title': 'Root1'},
+            'Root2': {'$ref': '#/$defs/Model2', 'title': 'Root2'},
         },
         'properties': {
             'a': {
@@ -2732,3 +2761,13 @@ def test_eq_with_cached_property():
     obj1.cached
     # Ensure the objects still compare equals after caching a property
     assert obj1 == obj2
+
+
+def test_model_metaclass_on_other_class() -> None:
+    """Test that `ModelMetaclass` can be used as a metaclass on an unrelated class.
+
+    This is done by some libraries to offer compatibility between Pydantic versions.
+    """
+
+    class OtherClass(metaclass=ModelMetaclass):
+        pass
